from typing import Any, Literal

from pydantic import Field, computed_field, field_validator, model_validator

from ...constants import HSR_LIGHT_CONE_RARITY_MAP
from ...enums import HSRPath
from ..base import APIModel

__all__ = ("LightCone", "LightConeDetail", "SuperimposeInfo")


class SuperimposeInfo(APIModel):
    """Light cone's superimpose information."""

    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    parameters: dict[str, list[float]] = Field(alias="Level")

    @model_validator(mode="before")
    def _flatten_parameters(cls, values: dict[str, Any]) -> dict[str, Any]:
        level = values["Level"]
        for key in level:
            level[key] = level[key]["ParamList"]

        return values


class LightConeDetail(APIModel):
    """HSR light cone detail."""

    id: int = Field(alias="Id")
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    path: HSRPath = Field(alias="BaseType")
    rarity: Literal[3, 4, 5] = Field(alias="Rarity")
    superimpose_info: SuperimposeInfo = Field(alias="Refinements")
    ascension_stats: list[dict[str, Any]] = Field(alias="Stats")

    @field_validator("rarity", mode="before")
    def _convert_rarity(cls, value: str) -> Literal[3, 4, 5]:
        return HSR_LIGHT_CONE_RARITY_MAP[value]

    @model_validator(mode="before")
    def _extract_id(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Use a hacky way to extract LC ID from relic recommendation.

        I don't understand why the API doesn't have the LC ID in the response.
        """
        values["Id"] = values["Stats"][0]["EquipmentID"]
        return values

    @computed_field
    @property
    def icon(self) -> str:
        """Light cone's icon URL."""
        return f"https://api.hakush.in/hsr/UI/lightconemediumicon/{self.id}.webp"

    @computed_field
    @property
    def image(self) -> str:
        """Light cone's image URL."""
        return self.icon.replace("lightconemediumicon", "lightconemaxfigures")


class LightCone(APIModel):
    """HSR light cone."""

    id: int  # This field is not present in the API response.
    rarity: Literal[3, 4, 5] = Field(alias="rank")
    description: str = Field(alias="desc")
    path: HSRPath = Field(alias="baseType")
    names: dict[Literal["en", "cn", "kr", "jp"], str]
    name: str = Field(None)  # The value of this field is assigned in post processing.

    @computed_field
    @property
    def icon(self) -> str:
        """Light cone's icon URL."""
        return f"https://api.hakush.in/hsr/UI/lightconemediumicon/{self.id}.webp"

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/hsr/UI/avatarshopicon/{value}.webp"

    @field_validator("rarity", mode="before")
    def _convert_rarity(cls, value: str) -> Literal[3, 4, 5]:
        return HSR_LIGHT_CONE_RARITY_MAP[value]

    @model_validator(mode="before")
    def _transform_names(cls, values: dict[str, Any]) -> dict[str, Any]:
        # This is probably the most questionable API design decision I've ever seen.
        values["names"] = {
            "en": values.pop("en"),
            "cn": values.pop("cn"),
            "kr": values.pop("kr"),
            "jp": values.pop("jp"),
        }
        return values



from typing import Any, Literal

from pydantic import Field, computed_field, field_validator, model_validator

from ...constants import HSR_CHARA_RARITY_MAP
from ...enums import HSRElement, HSRPath
from ..base import APIModel

__all__ = ("Character", "CharacterDetail", "Eidolon", "Skill", "SkillLevelInfo")


class SkillLevelInfo(APIModel):
    """Skill's level info."""

    level: int = Field(alias="Level")
    parameters: list[float] = Field(alias="ParamList")


class Skill(APIModel):
    """Character's skill."""

    name: str = Field(alias="Name")
    description: str | None = Field(None, alias="Desc")
    type: str | None = Field(None, alias="Type")
    tag: str = Field(alias="Tag")
    energy_generation: int | None = Field(None, alias="SPBase")
    level_info: dict[str, SkillLevelInfo] = Field(alias="Level")

    @computed_field
    @property
    def max_level(self) -> int:
        """Skill's max level."""
        return max(int(level) for level in self.level_info)


class Eidolon(APIModel):
    """Character's eidolon."""

    id: int = Field(alias="Id")
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    parameters: list[float] = Field(alias="ParamList")

    @computed_field
    @property
    def image(self) -> str:
        """Eidolon's image URL."""
        character_id = str(self.id)[:4]
        eidolon_index = str(self.id)[-1]
        return f"https://api.hakush.in/hsr/UI/rank/_dependencies/textures/{character_id}/{character_id}_Rank_{eidolon_index}.webp"


class CharacterDetail(APIModel):
    """HSR character detail."""

    id: int = Field(alias="Id")
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    rarity: Literal[4, 5] = Field(alias="Rarity")
    eidolons: dict[str, Eidolon] = Field(alias="Ranks")
    skills: dict[str, Skill] = Field(alias="Skills")
    ascension_stats: dict[str, dict[str, Any]] = Field(alias="Stats")

    @field_validator("rarity", mode="before")
    def _convert_rarity(cls, value: str) -> Literal[4, 5]:
        return HSR_CHARA_RARITY_MAP[value]

    @model_validator(mode="before")
    def _extract_id(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Use a hacky way to extract character ID from relic recommendation.

        I don't understand why the API doesn't have the character ID in the response.
        """
        values["Id"] = values["Relics"]["AvatarID"]
        return values

    @property
    def icon(self) -> str:
        """Character's icon URL."""
        return f"https://api.hakush.in/hsr/UI/avatarshopicon/{self.id}.webp"

    @property
    def gacha_art(self) -> str:
        """Character's gacha art URL."""
        return self.icon.replace("avatarshopicon", "avatardrawcard")


class Character(APIModel):
    """HSR character."""

    id: int  # This field is not present in the API response.
    icon: str
    rarity: Literal[4, 5] = Field(alias="rank")
    description: str = Field(alias="desc")
    path: HSRPath = Field(alias="baseType")
    element: HSRElement = Field(alias="damageType")
    names: dict[Literal["en", "cn", "kr", "jp"], str]
    name: str = Field(None)  # The value of this field is assigned in post processing.

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/hsr/UI/avatarshopicon/{value}.webp"

    @field_validator("rarity", mode="before")
    def _convert_rarity(cls, value: str) -> Literal[4, 5]:
        return HSR_CHARA_RARITY_MAP[value]

    @model_validator(mode="before")
    def _transform_names(cls, values: dict[str, Any]) -> dict[str, Any]:
        # This is probably the most questionable API design decision I've ever seen.
        values["names"] = {
            "en": values.pop("en"),
            "cn": values.pop("cn"),
            "kr": values.pop("kr"),
            "jp": values.pop("jp"),
        }
        return values



from typing import Any, Literal

from pydantic import Field, field_validator, model_validator

from ..base import APIModel

__all__ = (
    "Artifact",
    "ArtifactSet",
    "ArtifactSetDetail",
    "ArtifactSetDetailSetEffects",
    "ArtifactSetEffect",
    "ArtifactSetEffects",
    "SetEffect",
)


class SetEffect(APIModel):
    """Artifact set's set effect."""

    id: int
    affix_id: int = Field(alias="affixId")
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    parameters: list[float] = Field(alias="paramList")


class ArtifactSetDetailSetEffects(APIModel):
    """Artifact set's set effects."""

    two_piece: SetEffect
    four_piece: SetEffect | None = None


class Artifact(APIModel):
    """Genshin Impact artifact."""

    icon: str = Field(alias="Icon")
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"


class ArtifactSetDetail(APIModel):
    """Genshin Impact artifact set detail."""

    id: int = Field(alias="Id")
    icon: str = Field(alias="Icon")
    set_effect: ArtifactSetDetailSetEffects = Field(alias="Affix")
    parts: dict[str, Artifact] = Field(alias="Parts")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"

    @field_validator("set_effect", mode="before")
    def _assign_set_effect(cls, value: list[dict[str, Any]]) -> dict[str, Any]:
        return {
            "two_piece": value[0],
            "four_piece": value[1] if len(value) > 1 else None,
        }


class ArtifactSetEffect(APIModel):
    """Artifact set effect."""

    names: dict[Literal["EN", "KR", "CHS", "JP"], str]
    name: str = Field(None)  # The value of this field is assigned in post processing.
    descriptions: dict[Literal["EN", "KR", "CHS", "JP"], str] = Field(alias="desc")
    description: str = Field(None)  # The value of this field is assigned in post processing.

    @model_validator(mode="before")
    def _transform_names(cls, values: dict[str, Any]) -> dict[str, Any]:
        values["names"] = values.pop("name")
        return values


class ArtifactSetEffects(APIModel):
    """Artifact set effects."""

    two_piece: ArtifactSetEffect
    four_piece: ArtifactSetEffect | None = None


class ArtifactSet(APIModel):
    """Genshin Impact artifact set."""

    id: int
    icon: str
    rarities: list[int] = Field(alias="rank")
    set_effect: ArtifactSetEffects = Field(alias="set")
    names: dict[Literal["EN", "KR", "CHS", "JP"], str]
    name: str = Field(None)  # The value of this field is assigned in post processing.

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"

    @field_validator("set_effect", mode="before")
    def _assign_set_effects(cls, value: dict[str, Any]) -> dict[str, Any]:
        return {
            "two_piece": next(iter(value.values())),
            "four_piece": list(value.values())[1] if len(value) > 1 else None,
        }

    @model_validator(mode="before")
    def _extract_names(cls, values: dict[str, Any]) -> dict[str, Any]:
        values["names"] = next(iter(values["set"].values()))["name"]
        return values



from pydantic import Field

from ..base import APIModel

__all__ = ("DriveDisc", "DriveDiscDetail", "DriveDiscInfo")


class DriveDiscInfo(APIModel):
    """ZZZ drive disc info model."""

    name: str
    two_piece_effect: str = Field(alias="desc2")
    four_piece_effect: str = Field(alias="desc4")


class DriveDisc(APIModel):
    """ZZZ drive disc model."""

    id: int
    icon: str
    name: str = Field(None)  # This field doesn't exist in the API response
    two_piece_effect: str = Field(None)  # Same here
    four_piece_effect: str = Field(None)  # Same here

    en_info: DriveDiscInfo = Field(alias="EN")
    ko_info: DriveDiscInfo = Field(alias="KO")
    chs_info: DriveDiscInfo = Field(alias="CHS")
    ja_info: DriveDiscInfo = Field(alias="JA")


class DriveDiscDetail(APIModel):
    """ZZZ drive disc detail model."""

    id: int = Field(alias="Id")
    name: str = Field(alias="Name")
    two_piece_effect: str = Field(alias="Desc2")
    four_piece_effect: str = Field(alias="Desc4")
    story: str = Field(alias="Story")
    icon: str = Field(alias="Icon")
    icon2: str = Field(alias="Icon2")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, overload

from hakushin.enums import Game, Language

from .clients import GIClient, HSRClient, ZZZClient

if TYPE_CHECKING:
    from aiohttp import ClientSession

__all__ = ("HakushinAPI",)


class HakushinAPI:
    """Client to interact with the Hakushin API."""

    @overload
    def __new__(
        cls,
        game: Literal[Game.GI],
        lang: Language = Language.EN,
        *,
        cache_path: str = "./.cache/hakushin/aiohttp-cache.db",
        cache_ttl: int = 3600,
        headers: dict[str, Any] | None = None,
        debug: bool = False,
        session: ClientSession | None = None,
    ) -> GIClient: ...
    @overload
    def __new__(
        cls,
        game: Literal[Game.HSR],
        lang: Language = Language.EN,
        *,
        cache_path: str = "./.cache/hakushin/aiohttp-cache.db",
        cache_ttl: int = 3600,
        headers: dict[str, Any] | None = None,
        debug: bool = False,
        session: ClientSession | None = None,
    ) -> HSRClient: ...
    @overload
    def __new__(
        cls,
        game: Literal[Game.ZZZ],
        lang: Language = Language.EN,
        *,
        cache_path: str = "./.cache/hakushin/aiohttp-cache.db",
        cache_ttl: int = 3600,
        headers: dict[str, Any] | None = None,
        debug: bool = False,
        session: ClientSession | None = None,
    ) -> ZZZClient: ...
    def __new__(
        cls,
        game: Game,
        lang: Language = Language.EN,
        *,
        cache_path: str = "./.cache/hakushin/aiohttp-cache.db",
        cache_ttl: int = 3600,
        headers: dict[str, Any] | None = None,
        debug: bool = False,
        session: ClientSession | None = None,
    ) -> GIClient | HSRClient | ZZZClient:
        """Initializes the Hakushin API client."""
        if game is Game.GI:
            return GIClient(
                lang,
                cache_path=cache_path,
                cache_ttl=cache_ttl,
                headers=headers,
                debug=debug,
                session=session,
            )
        if game is Game.HSR:
            return HSRClient(
                lang,
                cache_path=cache_path,
                cache_ttl=cache_ttl,
                headers=headers,
                debug=debug,
                session=session,
            )
        return ZZZClient(
            lang,
            cache_path=cache_path,
            cache_ttl=cache_ttl,
            headers=headers,
            debug=debug,
            session=session,
        )



from pydantic import Field, computed_field

from ..base import APIModel

__all__ = ("ZZZExtraProp", "ZZZMaterial")


class ZZZMaterial(APIModel):
    """Generic ZZZ material."""

    id: int
    amount: int


class ZZZExtraProp(APIModel):
    """Generic ZZZ extra property."""

    id: int = Field(alias="Prop")
    name: str = Field(alias="Name")
    format: str = Field(alias="Format")
    value: int = Field(alias="Value")

    @computed_field
    @property
    def formatted_value(self) -> str:
        """Formatted value of this prop."""
        if "%" in self.format:
            return f"{self.value / 100:.0%}%"
        return str(self.value)



from pydantic import Field

from ..base import APIModel

__all__ = ("New",)


class New(APIModel):
    """ZZZ new stuff."""

    character_ids: list[int] = Field(alias="character")
    bangboo_ids: list[int] = Field(alias="bangboo")
    weapon_ids: list[int] = Field(alias="weapon")
    equipment_ids: list[int] = Field(alias="equipment")
    item_ids: list[int] = Field(alias="item")
    current_version: str = Field(alias="version")
    previous_versions: list[str] = Field(alias="previous")
